# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provide implement for pipeline.
"""
import logging
from pymochow.ai.pipeline import Pipeline
from typing import Any, Dict, List
from pymochow.ai.embedder import Embedder
from pymochow.model.table import (
    Table,
    FloatVector,
    VectorSearchRequest,
    VectorTopkSearchRequest,
    VectorRangeSearchRequest,
    VectorBatchSearchRequest,
    BM25SearchRequest,
    HybridSearchRequest
)
from pymochow.model.enum import ReadConsistency

_logger = logging.getLogger(__name__)

class DefaultPipeline(Pipeline):
    """
    Implement for Pipeline class.
    """
    def __init__(self, batch_size=100):
        """
        Initialize the DefaultPipeline instance.
        """
        self._batch_size = batch_size
    
    def ingest_doc(
        self,
        doc, 
        doc_processor=None, 
        embedder=None, 
        meta_table=None,
        doc_to_row_mapping=None,
        chunk_table=None,
        chunk_to_row_mapping=None
    ):
        """
        Process and store a document and its chunks into the specified database tables.
        
        Parameters:
            doc (Document): The document object to be ingested.
            doc_processor (Processor, optional): A tool to parse and split document. 
            embedder (Embedder, optional): A tool to generate embeddings of the document content.
            meta_table (Table, optional): The table in the database where the document data will be stored.
            doc_to_row_mapping (dict, optional): A dictionary that maps document object attributes to 
                database table columns for the document.
            chunk_table (Table, optional): The table in the database used to store document chunks.
            chunk_to_row_mapping (dict, optional): A dictionary that maps chunk object attributes to 
                database table columns for the chunks.
        """
        if doc_processor is None:
            raise ValueError("doc_processor must be provided and non-None")
        if embedder is None:
            raise ValueError("embedder must be provided and non-None")
        if meta_table is None:
            raise ValueError("meta_table must be provided and non-None")
        if chunk_table is None:
            raise ValueError("chunk_table must be provided and non-None")

        rows = []
        rows.append(doc.to_row(doc_to_row_mapping))
        meta_table.insert(rows=rows)

        rows = []
        chunks = embedder.embedding(doc_processor.process_doc(doc=doc))
        for chunk in chunks:
            rows.append(chunk.to_row(chunk_to_row_mapping))
        for i in range(0, len(rows), self._batch_size):
            batch_rows = rows[i:i + self._batch_size]
            chunk_table.insert(rows=batch_rows)

    def vector_search(
        self, 
        search_contents: List[str], 
        embedder: Embedder, 
        table: Table, 
        search_request: VectorSearchRequest,
        partition_key: Dict[str, Any] = None,
        projections: List[str] = None,
        read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
        config: Dict[Any, Any] = None
    ):
        """
        Perform a vector-based search operation.

        This method converts the search content into a vector using the embedder and performs a
        vector search in the specified table, returning the most similar results to the query.

        Parameters:
        ----------
        search_contents : List[str]
            The input search content, usually in the form of text or already embedded vector.
        embedder : Embedder
            The embedder object used to convert the search content into vector form.
        table : Table
            The target table where the search is conducted.
        search_request : VectorSearchRequest
            The search request object containing parameters like TopK, filters, etc.
        partition_key : Dict[str, Any], optional
            The partition key to narrow the search to a specific partition, default is None.
        projections : List[str], optional
            The list of fields to include in the search result, default is None.
        read_consistency : ReadConsistency, optional
            The level of read consistency required, default is EVENTUAL.
        config : Dict[Any, Any], optional
            Additional configurations for the search, default is None.
        """
        if not search_contents or len(search_contents) == 0:
            raise ValueError("search_contents must not be None or empty")

        embeddings = embedder.embedding_text(search_contents)
        _logger.debug("embeddings result: {}".format(embeddings))

        if len(search_contents) == 1:
            if not isinstance(search_request, (VectorTopkSearchRequest, VectorRangeSearchRequest)):
                raise ValueError("For a single search content, search_request must be \
                        VectorTopkSearchRequest or VectorRangeSearchRequest")
            # embeddings[0] is the first and only embedding
            search_request.vector = FloatVector(embeddings[0])

        else:
            if not isinstance(search_request, VectorBatchSearchRequest):
                raise ValueError("For multiple search contents, search_request must be VectorBatchSearchRequest")

            search_request.vectors = [FloatVector(embedding) for embedding in embeddings]
        
        _logger.debug("search_request: {}".format(search_request.to_dict()))

        result = table.vector_search(
            request=search_request,
            partition_key=partition_key,
            projections=projections,
            read_consistency=read_consistency,
            config=config
        )

        return result

    def bm25_search(
        self,
        table: Table,
        search_request: BM25SearchRequest,
        partition_key: Dict[str, Any] = None,
        projections: List[str] = None,
        read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
        config: Dict[Any, Any] = None
    ):
        """
        Perform a BM25-based text search operation.

        This method uses the BM25 search algorithm to perform a text-based search on the specified
        table, returning the most relevant documents based on the query terms.

        Parameters:
        ----------
        table : Table
            The target table where the search will be performed.
        search_request : BM25SearchRequest
            The search request object, which contains query terms and other parameters.
        partition_key : Dict[str, Any], optional
            The partition key to narrow the search to a specific partition, default is None.
        projections : List[str], optional
            The list of fields to include in the search result, default is None.
        read_consistency : ReadConsistency, optional
            The level of read consistency required, default is EVENTUAL.
        config : Dict[Any, Any], optional
            Additional configurations for the search, default is None.
        """
        result = table.bm25_search(
            request=search_request, 
            partition_key=partition_key,
            projections=projections,
            read_consistency=read_consistency,
            config=config
        )
        return result

    def hybrid_search(
        self, 
        search_contents: List[str], 
        embedder: Embedder, 
        table: Table, 
        search_request: HybridSearchRequest,
        partition_key: Dict[str, Any] = None,
        projections: List[str] = None,
        read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
        config: Dict[Any, Any] = None
    ):
        """
        Perform a hybrid search (vector + traditional text search).

        This method combines vector search with BM25 text search, suitable for scenarios
        requiring both semantic and keyword-based search.

        Parameters:
        ----------
        search_contents : List[str]
            The input search content, usually text-based.
        embedder : Embedder
            The embedder object used to convert the search content into vectors.
        table : Table
            The target table where the search is conducted.
        partition_key : Dict[str, Any], optional
            The partition key to narrow the search to a specific partition, default is None.
        projections : List[str], optional
            The list of fields to include in the search result, default is None.
        read_consistency : ReadConsistency, optional
            The level of read consistency required, default is EVENTUAL.
        config : Dict[Any, Any], optional
            Additional configurations for the search, default is None.
        """
        if not search_contents or len(search_contents) == 0:
            raise ValueError("search_contents must not be None or empty")
        
        embeddings = embedder.embedding_text(search_contents)

        if len(search_contents) == 1:
            if not isinstance(search_request._vector_request, (VectorTopkSearchRequest, VectorRangeSearchRequest)):
                raise ValueError("For a single search content, search_request \
                        must be VectorTopkSearchRequest or VectorRangeSearchRequest")
            # embeddings[0] is the first and only embedding
            search_request._vector_request.vector = FloatVector(embeddings[0])

        else:
            if not isinstance(search_request._vector_request, VectorBatchSearchRequest):
                raise ValueError("For multiple search contents, search_request must be VectorBatchSearchRequest")

            search_request._vector_request.vectors = [FloatVector(embedding) for embedding in embeddings]

        result = table.hybrid_search(
            request=search_request,
            partition_key=partition_key,
            projections=projections,
            read_consistency=read_consistency,
            config=config
        )

        return result



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide interface for doc parser.
"""
from pymochow.model.document import DocumentChunk
from abc import ABC, abstractmethod
from typing import List


class DocParser(ABC):
    """
    DocParser: An abstract base class that defines the interface for parsing  documents into chunks.

    Subclasses must implement the `process_doc` method to parse the document based on specific criteria,
    such as page length, overlap length, and the number of pages to take.
    """
    @abstractmethod
    def parse(self, doc) -> DocumentChunk:
        """
        Parse document into chunks based on the provided parameters.

        Parameters:
        ----------
        doc : Document
            The document to be parsed and split into chunks.
        
        Returns:
        -------
        DocumentChunk
            A DocumentChunk object.
        """
        pass



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide interface for doc parser.
"""
from pymochow.model.enum import DocSplitMode
from pymochow.model.document import DocumentChunk
from abc import ABC, abstractmethod
from typing import List


class DocProcessor(ABC):
    """
    DocProcessor: An abstract base class that defines the interface for parsing and splitting documents into chunks.

    Subclasses must implement the `process_doc` method to parse and split the document based on specific criteria,
    such as page length, overlap length, and the number of pages to take.
    """
    @abstractmethod
    def process_doc(self, doc) -> List[DocumentChunk]:
        """
        Parse and split the document into chunks based on the provided parameters.

        Parameters:
        ----------
        doc : Document
            The document to be parsed and split into chunks.
        
        Returns:
        -------
        List[DocumentChunk]
            A list of DocumentChunk objects, where each chunk represents a part of the parsed document.
        """
        pass



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide interface for doc hub.
"""
from pymochow.model.document import Document
from abc import ABC, abstractmethod
from typing import List

# Abstract base class DocumentHub, defines the basic interface for a document hub
class DocumentHub(ABC):
    """
    DocumentHub: Abstract base class for the document hub that defines the basic operations.
    Subclasses should implement methods for adding, removing, listing, and downloading documents.
    """

    @abstractmethod
    def add(self, doc: Document) -> Document:
        """
        Add a document to the hub.

        Parameters:
        doc (Document): The document to be added.
        """
        pass

    @abstractmethod
    def remove(self, doc: Document):
        """
        Remove a document from the hub.

        Parameters:
        doc (Document): The document to be removed.
        """
        pass

    @abstractmethod
    def list(self) -> List[Document]:
        """
        List all documents available in the hub.

        Returns:
        List[Document]: A list of all available document objects.
        """
        pass

    @abstractmethod
    def load(self, doc: Document) -> Document:
        """
        Load a document from the hub.

        Parameters:
        doc (Document): The document to load.

        Returns:
        doc (Document): If the document in hub then load it.
        """
        pass


# Environment configuration class for DocumentHub
class DocumentHubEnv:
    """
    DocumentHubEnv: Stores configuration settings for interacting with an object storage service
    or a local file system. This class holds details like endpoint URLs, authentication keys,
    work path for object storage, and a local cache path.

    Parameters:
    ----------
    endpoint (str):
        The endpoint URL for the object storage service, such as an S3 or BOS-compatible service.
    ak (str):
        Access key for authentication with the object storage service.
    sk (str):
        Secret key for authentication with the object storage service.
    root_path (str):
        The root path where documents are stored. For object storage, this should be in the format
        'bos://bucket_name/object_prefix'. For local file systems, this should be in the format
        'local://root_path/'.
    local_cache_path (str):
        The local file path where downloaded documents will be cached. This parameter is only relevant
        when documents are stored in object storage (like BOS). It defines the location where documents
        are cached for local access.
    """

    def __init__(self, endpoint=None, ak=None, sk=None,
            root_path=None, local_cache_path=None):
        self._endpoint = endpoint
        self._access_key = ak
        self._secret_key = sk
        self._root_path = root_path
        self._local_cache_path = local_cache_path

    def endpoint(self):
        """
        Get the endpoint URL for the object storage service.

        Returns:
        str: The endpoint URL.
        """
        return self._endpoint

    def access_key(self):
        """
        Get the access key for the object storage service.

        Returns:
        str: The access key.
        """
        return self._access_key

    def secret_key(self):
        """
        Get the secret key for the object storage service.

        Returns:
        str: The secret key.
        """
        return self._secret_key

    def root_path(self):
        """
        Get the root path where documents are stored. 

        Returns:
        -------
        str: 
            The root path. For object storage, it would be in the format 'bos://bucket_name/object_prefix'. 
            For local file systems, it would be in the format 'local://root_path/'.
        """
        return self._root_path

    def local_cache_path(self):
        """
        Get the local cache path where downloaded documents are stored for quick access.
        This is only used when the documents are stored in object storage.

        Returns:
        -------
        str:
            The local file path where documents will be cached.
        """
        return self._local_cache_path



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide local document hub.
"""
from pymochow.model.document import Document
from pymochow.ai.dochub import DocumentHub, DocumentHubEnv
import logging
import os
import shutil
import time
from datetime import datetime
from typing import List
from baidubce.services.bos.bos_client import BosClient
from baidubce.bce_client_configuration import BceClientConfiguration
from baidubce.auth.bce_credentials import BceCredentials
from baidubce.exception import BceHttpClientError

_logger = logging.getLogger(__name__)

class BosDocumentHub(DocumentHub):
    """
    BOSDocumentHub: A concrete implementation of DocumentHub that manages documents in Baidu Object Storage (BOS).
    """

    def __init__(self, env: DocumentHubEnv):
        """
        Initialize the BOSDocumentHub with credentials and region information.
        """
        self._env = env
        if not env.endpoint() or not env.access_key() or not env.secret_key() or not env.local_cache_path():
            raise ValueError("All parameters (endpoint, access key, secret key, local cache path) must be provided.")
        
        self._schema = "bos://"

        # Extract the bucket name and object prefix from the work path
        if env.root_path().startswith(self._schema):
            path_parts = env.root_path()[6:].split('/', 1)  # Strip 'bos://' and split into bucket and prefix
            self._bucket_name = path_parts[0]
            self._object_prefix = path_parts[1] if len(path_parts) > 1 else ''
        else:
            raise ValueError("Invalid root_path format for BOSDocumentHub. Expected 'bos://bucket_name/object_prefix'.")

        # Create BOS client
        self._client = self._create_bos_client(env.endpoint(), env.access_key(), env.secret_key())
    
    def _create_bos_client(self, endpoint, access_key, secret_key):
        """
        Create and configure the BOS client with the given credentials and endpoint.

        Parameters:
        endpoint (str): BOS service endpoint.
        access_key (str): Access key for BOS authentication.
        secret_key (str): Secret key for BOS authentication.

        Returns:
        BosClient: Configured client for interacting with Baidu Object Storage.
        """
        credentials = BceCredentials(access_key, secret_key)
        config = BceClientConfiguration(credentials, endpoint)
        return BosClient(config)
    
    def add(self, doc: Document) -> Document:
        """
        Upload a document to the BOS bucket.
        """
        object_key = f"{self._object_prefix}/{doc.doc_name}" if self._object_prefix else doc.doc_name
        self._client.put_object_from_file(self._bucket_name, object_key, doc.file_path)
        doc.uri = f"{self._schema}{self._bucket_name}/{object_key}"
        doc.size = os.path.getsize(doc.file_path)  # Get file size
        doc.ctime = int(time.time())
        return doc

    def remove(self, doc: Document):
        """
        Delete a document from the BOS bucket.
        """
        object_key = f"{self._object_prefix}/{doc.doc_name}" if self._object_prefix else doc.doc_name
        try:
            self._client.delete_object(self._bucket_name, object_key)
        except BceHttpClientError as e:
            if e.code == 'NoSuchKey':  # Check if the error code corresponds to a missing object
                _logger.info(f"Document '{doc.doc_name}' does not exist. No action needed.")
            else:
                raise  # Re-raise the exception if it is not a 'NoSuchKey' error
        except Exception as e:
            _logger.warning(f"Unexpected error occurred while deleting '{doc.doc_name}': {e}")
            raise

    def list(self) -> List[Document]:
        """
        List all documents stored in the BOS bucket.
        """
        response = self._client.list_objects(self._bucket_name, prefix=self._object_prefix)
        documents = []
        for obj in response.contents:
            if obj.key.endswith('/'):  # Skip directories
                continue
            doc_name = obj.key[len(self._object_prefix) + 1:] if self._object_prefix else obj.key
            doc_uri = f"{self._schema}{self._bucket_name}/{obj.key}"
            ctime = int(datetime.strptime(obj.last_modified, "%Y-%m-%dT%H:%M:%SZ").timestamp())
            
            documents.append(Document(
                doc_name=doc_name,
                uri=doc_uri,
                size=obj.size,  # Size in bytes
                ctime=ctime  # Creation time as a timestamp
            ))
        return documents

    def load(self, doc: Document) -> Document:
        """
        Load a document if it exists in the BOS bucket.
        """
        object_key = f"{self._object_prefix}/{doc.doc_name}" if self._object_prefix else doc.doc_name
        local_file_path = os.path.join(self._env.local_cache_path(), doc.doc_name)
        os.makedirs(os.path.dirname(local_file_path), exist_ok=True)
        
        response = self._client.get_object_meta_data(self._bucket_name, object_key)
        doc.size = response.metadata.content_length  # 更新 size
        doc.ctime = int(datetime.strptime(
            response.metadata.last_modified, "%a, %d %b %Y %H:%M:%S GMT"
        ).timestamp())

        self._client.get_object_to_file(self._bucket_name, object_key, local_file_path)
        doc.file_path = local_file_path
        doc.uri = f"{self._schema}{self._bucket_name}/{object_key}"
        return doc



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provide schema.
"""
from typing import Dict, List
from pymochow.model.enum import (
    IndexType, AutoBuildPolicyType, InvertedIndexAnalyzer, InvertedIndexParseMode, InvertedIndexFieldAttribute
)


class AutoBuildTiming:
    """
    AutoBuildTiming
    """

    def __init__(self, timing):
        self._timing = timing
        self._auto_build_policy_type = AutoBuildPolicyType.TIMING

    def to_dict(self):
        """to dict"""
        res = {"policyType": self._auto_build_policy_type, "timing": self._timing}
        return res


class AutoBuildPeriodical:
    """
    AutoBuildPeriodical
    """

    def __init__(self, period_s, timing=""):
        """
        __init__
        """
        self._period_s = period_s
        self._auto_build_policy_type = AutoBuildPolicyType.PERIODICAL
        self._timing = timing

    def to_dict(self):
        """to dict"""
        res = {
            "policyType": self._auto_build_policy_type,
            "periodInSecond": self._period_s,
        }
        if self._timing != "":
            res["timing"] = self._timing
        return res


class AutoBuildRowCountIncrement:
    """
    AutoBuildRowCountIncrement
    """

    def __init__(self, row_count_increment=0, row_count_increment_ratio=0):
        self._row_count_increment = row_count_increment
        self._row_count_increment_ratio = row_count_increment_ratio
        self._auto_build_policy_type = AutoBuildPolicyType.ROW_COUNT_INCREMENT

    def to_dict(self):
        """to dict"""
        res = {
            "policyType": self._auto_build_policy_type,
        }
        if self._row_count_increment != 0:
            res["rowCountIncrement"] = self._row_count_increment
        if self._row_count_increment_ratio != 0:
            res["rowCountIncrementRatio"] = self._row_count_increment_ratio
        return res


class AutoBuildTool:
    """
    AutoBuildTool
    """

    def get_auto_build_index_policy(auto_build_policy_dict):
        """get auto build index policy"""
        if "policyType" not in auto_build_policy_dict:
            return None
        if (auto_build_policy_dict["policyType"].upper() == AutoBuildPolicyType.TIMING.name):
            return AutoBuildTiming(auto_build_policy_dict["timing"])
        elif (auto_build_policy_dict["policyType"].upper() == AutoBuildPolicyType.PERIODICAL.name):
            if "timing" in auto_build_policy_dict:
                return AutoBuildPeriodical(auto_build_policy_dict["periodInSecond"], auto_build_policy_dict["timing"])
            else:
                return AutoBuildPeriodical(auto_build_policy_dict["periodInSecond"])
        elif (auto_build_policy_dict["policyType"].upper() == AutoBuildPolicyType.ROW_COUNT_INCREMENT.name):
            row_count_increment = 0
            row_count_increment_ratio = 0.0
            if "rowCountIncrementRatio" in auto_build_policy_dict:
                row_count_increment_ratio = auto_build_policy_dict["rowCountIncrementRatio"]
            if "rowCountIncrement" in auto_build_policy_dict:
                row_count_increment = auto_build_policy_dict["rowCountIncrement"]
            return AutoBuildRowCountIncrement(row_count_increment, row_count_increment_ratio)

DefaultAutoBuildPeriodical = 24 * 3600
DefaultAutoBuildPolicy = AutoBuildPeriodical(DefaultAutoBuildPeriodical)

class Field:
    """field"""

    def __init__(self,
            field_name,
            field_type,
            primary_key=False,
            partition_key=False,
            auto_increment=False,
            not_null=False,
            dimension=0):
        self._field_name = field_name
        self._field_type = field_type
        self._primary_key = primary_key
        self._partition_key = partition_key
        self._auto_increment = auto_increment
        self._not_null = not_null
        self._dimension = dimension

    @property
    def field_name(self):
        """field name"""
        return self._field_name

    @property
    def field_type(self):
        """field type"""
        return self._field_type

    @property
    def primary_key(self):
        """primary key"""
        return self._primary_key

    @property
    def partition_key(self):
        """partition key"""
        return self._partition_key

    @property
    def auto_increment(self):
        """auto increment"""
        return self._auto_increment

    @property
    def not_null(self):
        """not null"""
        return self._not_null

    @property
    def dimension(self):
        """dimension"""
        return self._dimension

    def to_dict(self):
        """to dict"""
        res = {
            "fieldName": self.field_name,
            "fieldType": self.field_type,
            "notNull": self.not_null
        }
        if self.primary_key:
            res["primaryKey"] = True
        if self.partition_key:
            res["partitionKey"] = True
        if self.auto_increment:
            res["autoIncrement"] = True
        if self.dimension > 0:
            res["dimension"] = self.dimension
        return res


class IndexField:
    """index field"""

    def __init__(self, index_name, field, index_type):
        self._index_name = index_name
        self._field = field
        self._index_type = index_type

    @property
    def index_type(self):
        """index type"""
        return self._index_type

    @property
    def index_name(self):
        """index name"""
        return self._index_name

    @property
    def field(self):
        """field"""
        return self._field


class HNSWParams:
    """
    The hnsw vector index params.
    """

    def __init__(self, m: int, efconstruction: int) -> None:
        self.m = m
        self.ef_construction = efconstruction

    def to_dict(self):
        """to dict"""
        res = {
            "M": self.m,
            "efConstruction": self.ef_construction
        }
        return res


class HNSWPQParams:
    """
    The hnsw pq vector index params.
    """

    def __init__(self, m: int, efconstruction: int, NSQ: int, samplerate: float) -> None:
        self.m = m
        self.ef_construction = efconstruction
        self.NSQ = NSQ
        self.samplerate = samplerate
    
    def to_dict(self):
        """to dict"""
        res = {
            "M": self.m,
            "efConstruction": self.ef_construction,
            "NSQ": self.NSQ,
            "sampleRate": self.samplerate
        }
        return res


class PUCKParams:
    """
    The puck vector index params.
    """

    def __init__(self, coarseClusterCount: int, fineClusterCount: int) -> None:
        self.coarseClusterCount = coarseClusterCount
        self.fineClusterCount = fineClusterCount

    def to_dict(self):
        """to dict"""
        res = {
            "coarseClusterCount": self.coarseClusterCount,
            "fineClusterCount": self.fineClusterCount
        }
        return res

    
class VectorIndex(IndexField):
    """
    Args:
        index_name(str): The field name of the index.
        field(str): make index on which field
        metric_type(MetricType): The metric type of the vector index.
        params(Any): HNSWParams if the index_type is HNSW
        auto_build(boolean): auto build vector index
    """
    def __init__(
            self,
            index_name,
            index_type,
            field,
            metric_type,
            params=None,
            auto_build=False,
            auto_build_index_policy=None,
            **kwargs):
        super().__init__(index_name=index_name, index_type=index_type, field=field)
        self._metric_type = metric_type
        self._params = params
        self._auto_build = auto_build
        if self._auto_build:
            self._auto_build_index_policy = auto_build_index_policy
        else:
            self._auto_build_index_policy = None
        self._state = kwargs.get('state', None)

    @property
    def metric_type(self):
        """metric type"""
        return self._metric_type

    @property
    def params(self):
        """params"""
        return self._params

    @property
    def auto_build(self):
        """auto build"""
        return self._auto_build

    @property
    def state(self):
        """state"""
        return self._state
    @property
    def auto_build_index_policy(self):
        """state"""
        return self._auto_build_index_policy

    def to_dict(self):
        """to dict"""
        res = {
            "indexName": self.index_name,
            "indexType": self.index_type,
            "field": self.field,
            "metricType": self.metric_type,
            "autoBuild": self.auto_build
        }
        if self.params is not None:
            res["params"] = self.params.to_dict()
        if self.state is not None:
            res["state"] = self.state
        if self.auto_build_index_policy is not None:
            res["autoBuildPolicy"] = self.auto_build_index_policy.to_dict()
        return res


class SecondaryIndex(IndexField):
    """secondary index"""

    def __init__(
            self,
            index_name,
            field):
        super().__init__(index_name=index_name, index_type=IndexType.SECONDARY_INDEX,
                field=field)

    def to_dict(self):
        """to dict"""
        res = {
            "indexName": self.index_name,
            "indexType": self.index_type,
            "field": self.field
        }
        return res


class InvertedIndexParams:
    """
    inverted index params
    """
    def __init__(
            self,
            analyzer: InvertedIndexAnalyzer=None,
            parse_mode: InvertedIndexParseMode=None):
        """init"""
        self._params = {}
        if analyzer is not None:
            self._params["analyzer"] = analyzer
        if parse_mode is not None:
            self._params["parseMode"] = parse_mode

    def to_dict(self) -> Dict[str, str]:
        """to dict"""
        return self._params


class InvertedIndex(IndexField):
    """inverted index (for BM25 search)"""
    def __init__(
            self,
            index_name: str,
            fields: List[str],
            params: InvertedIndexParams,
            field_attributes: List[InvertedIndexFieldAttribute] = []):
        """init

        InvertedIndex 用于在 create_table 时，为 'fields' 指定的列建立
        倒排索引。注意 fields 中包含的列应当是 TEXT、TEXT_GBK 或者
        TEXT_GB18030 类型，否则 create_table 会失败。

        'attributes' 可以为空, 否则, 应当与 fields 有相同数量的元素

        """
        super().__init__(index_name=index_name,
                         field=None,
                         index_type=IndexType.INVERTED_INDEX)
        self._fields = fields
        self._field_attributes = field_attributes
        self._params = params

    def to_dict(self):
        """to dict"""
        res = {
            "indexName": self.index_name,
            "indexType": self.index_type,
            "fields": self._fields,
            "params": self._params.to_dict()
        }
        if len(self._field_attributes) != 0:
            res["fieldsIndexAttributes"] = self._field_attributes
        return res


class Schema:
    """schema"""

    def __init__(self, fields, indexes=None):
        self._fields = fields
        self._indexes = indexes

    @property
    def indexes(self):
        """indexes"""
        return self._indexes

    @property
    def fields(self):
        """fields"""
        return self._fields

    def to_dict(self):
        """to dict"""
        res = {}
        res["fields"] = []
        for field in self.fields:
            res["fields"].append(field.to_dict())

        if self.indexes is not None:
            res["indexes"] = []
            for index in self.indexes:
                res["indexes"].append(index.to_dict())
        return res



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provide appbuilder implement to parser and split doc.
"""

import logging
from typing import List
from pymochow.ai.processor import DocProcessor
from pymochow.model.document import DocumentChunk
from pymochow.model.enum import (
    DocSplitMode,
    Lang
)
from pymochow.ai.parser import LangchainDocParser
from pymochow.ai.splitter import LangchainDocSplitter

_logger = logging.getLogger(__name__)

class LangchainDocProcessor(DocProcessor):
    """
    LangchainDocProcessor: A concrete implementation of the DocProcessor class. 
    This class is responsible to parser and split a document into chunks based on a given splitting mode.
    """

    def __init__(
        self, 
        split_mode=DocSplitMode.PAGE,
        maximum_page_length=800,
        page_overlap_length=200,
        maximum_pages_totake=None
    ):
        """
        Initialize the LangchainDocProcessor.
        split_mode : DocSplitMode
            The mode of splitting the document (e.g., by page). The default is to split by pages.
        
        maximum_page_length : int, optional
            The maximum length of a chunk in terms of 
            pages (or another unit depending on the split mode). Defaults to None.
        
        page_overlap_length : int, optional
            The number of overlapping units (pages, words, etc.) between consecutive chunks, 
            ensuring some content overlap. Defaults to None.
        
        maximum_pages_totake : int, optional
            The total number of pages (or another unit) to process from the document. 
            This acts as a limit on how much of the document is parsed. Defaults to None.
        """
        self._split_mode = split_mode
        self._maximum_page_length = maximum_page_length
        self._page_overlap_length = page_overlap_length
        self._maximum_pages_totake = maximum_pages_totake

    def process_doc(self, doc) -> List[DocumentChunk]:
        """
        Parse and split the document into chunks based on the provided parameters.
        
        Parameters:
        ----------
        doc : Document
            The document to be parsed and split into chunks.
        
        Returns:
        -------
        List[DocumentChunk]
            A list of DocumentChunk objects, where each chunk represents a part of the parsed document.
        """
        parser = LangchainDocParser()
        initial_chunk = parser.parse(doc)
        
        splitter = LangchainDocSplitter(
            split_mode=self._split_mode,
            maximum_page_length=self._maximum_page_length,
            page_overlap_length=self._page_overlap_length,
            maximum_pages_totake=self._maximum_pages_totake
        )
        chunks = splitter.split(initial_chunk)
        return chunks



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provide implement for qianfan embedding.
"""
import time
import logging
from typing import List
from pymochow.model.document import DocumentChunk
from pymochow.ai.embedder import Embedder
import appbuilder
from appbuilder import Message

_logger = logging.getLogger(__name__)

class QianfanEmbedder(Embedder):
    """
    QianfanEmbedder: Implementation of the Embedder class by Qianfan SDK. 
    """
    def __init__(self, model=None, batch=10, field_mapping=None):
        """
        Initialize the QianfanEmbedder.
        """
        if model is not None:
            self._embedding = appbuilder.Embedding(model)
        else:
            self._embedding = appbuilder.Embedding()

        if field_mapping is None:
            self._field_mapping = {'content': 'embedding'}
        else:
            self._field_mapping = field_mapping
        self._batch = batch

    def embedding(self, chunks) -> List[DocumentChunk]:
        """
        Generate embeddings for specified fields in document chunks.

        Parameters:
        chunks (List[DocumentChunk]): A list of document chunks that need to be processed for embeddings.
        Returns:
        List[DocumentChunk]: A list of `DocumentChunk` objects, 
            with the embeddings added to the corresponding fields based on `field_mapping`.
        """
        # Batch processing of document chunks for embeddings
        for i in range(0, len(chunks), self._batch):
            batch_chunks = chunks[i:i + self._batch]
            
            # Prepare data for each field that needs embedding
            for field, vector in self._field_mapping.items():
                field_data = [getattr(chunk, field) for chunk in batch_chunks]
                result = self._embedding.batch(Message(field_data))
                # Apply the embeddings to the corresponding vector field
                for chunk, embedding in zip(batch_chunks, result.content):
                    setattr(chunk, vector, embedding)

            time.sleep(1)  # Respect API rate limit

        return chunks
    
    def embedding_text(self, texts) -> List[List[float]]:
        """
        Processes a list of text strings and converts them into embeddings using a batch processing method.

        Parameters:
            texts (List[str]): A list of text strings to be embedded.

        Returns:
            List[List[float]]: A list of lists, where each inner list 
                is a numerical vector representing the embedded text.
        """
        embedding = []
        for i in range(0, len(texts), self._batch):
            batch_texts = texts[i:i + self._batch]
            result = self._embedding.batch(Message(batch_texts))
            embedding.extend(result.content)
        return embedding




# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide interface for embedding.
"""
from pymochow.model.document import DocumentChunk
from abc import ABC, abstractmethod
from typing import List

class Embedder(ABC):
    """
    Embedder: An abstract base class for generating embeddings for document chunks.
    """
    
    @abstractmethod
    def embedding(self, chunks) -> List[DocumentChunk]:
        """
        Generate embeddings for specified fields in document chunks.

        Parameters:
        chunks (List[DocumentChunk]): A list of document chunks that need to be processed for embeddings.
        Returns:
        List[DocumentChunk]: A list of `DocumentChunk` objects, 
            with the embeddings added to the corresponding fields based on `field_mapping`.
        """
        pass

    @abstractmethod
    def embedding_text(self, texts) -> List[List[float]]:
        """
        Embeds a given text into a numerical representation and returns the result.

        Args:
            text (List[str]): The text to be embedded.

        Returns:
            List[List[float]]: The numerical embedding of the text as a list of floats.

        Note:
            The method includes a sleep call to throttle the embedding rate due to API rate limits.
        """



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide table model.
"""
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union, Tuple
import copy
import orjson
from pymochow import utils
from pymochow import client
from pymochow.http import http_methods
from pymochow.model.schema import (
    VectorIndex,
    SecondaryIndex,
    HNSWParams,
    HNSWPQParams,
    PUCKParams,
    DefaultAutoBuildPolicy,
    AutoBuildTool,
    InvertedIndex,
    InvertedIndexParams,
    InvertedIndexAnalyzer,
    InvertedIndexParseMode
)
from pymochow.model.enum import (PartitionType, ReadConsistency,
    IndexType, IndexState, MetricType, AutoBuildPolicyType, RequestType)
from pymochow.exception import ClientError

class Partition:
    """
    Partition
    """
    def __init__(self, partition_num, partition_type=PartitionType.HASH):
        self._partition_num = partition_num
        self._partition_type = partition_type

    def to_dict(self):
        """to dict"""
        res = {
            "partitionType": self._partition_type,
            "partitionNum": self._partition_num
        }
        return res


@utils.deprecated("No longer used. Use SearchRequest instead.")
class AnnSearch:
    """ann search"""

    def __init__(self, vector_field, vector_floats, params, filter=None):
        """init"""
        self._vector_field = vector_field
        self._vector_floats = vector_floats
        self._params = params
        self._filter = filter

    def to_dict(self):
        """to dict"""
        res = {
            'vectorField': self._vector_field,
            'vectorFloats': self._vector_floats,
            'params': self._params.to_dict(),
        }
        if self._filter is not None:
            res['filter'] = self._filter
        return res


class Vector(ABC):
    """base class of vector"""
    @abstractmethod
    def representation(self):
        """representation"""
        pass


class BatchQueryKey:
    '''BatchQueryKey'''

    def __init__(self, primary_key, partition_key=None):
        '''init'''
        self._primary_key = primary_key
        self._partition_key = partition_key

    def to_dict(self):
        '''to dict'''

        res = {'primaryKey': self._primary_key}
        if self._partition_key is not None:
            res['partitionKey'] = self._partition_key
        return res


class FloatVector(Vector):
    """float vector"""

    def __init__(self, floats: List[float]):
        """init"""
        self._floats = floats

    def representation(self):
        """representation"""
        return self._floats


class VectorSearchConfig:
    """
    Optional configurable params for vector search.

    For each index algorithm, the params that could be set are:

    | IndexType | Params              |
    |-----------+---------------------|
    | HNSW      | ef, pruning         |
    | HNSWPQ    | ef, pruning         |
    | PUCK      | search_coarse_count |
    | FLAT      |                     |

    """

    def __init__(self, *,
                 ef: int = None,
                 pruning: bool = None,
                 search_coarse_count: int = None):
        """init"""
        self._ef = ef
        self._pruning = pruning
        self._search_coarse_count = search_coarse_count

    def to_dict(self):
        """to_dict"""
        res = {}
        if self._ef is not None:
            res['ef'] = self._ef
        if self._pruning is not None:
            res['pruning'] = self._pruning
        if self._search_coarse_count is not None:
            res['searchCoarseCount'] = self._search_coarse_count
        return res


class SearchRequest(ABC):
    """base class"""
    @abstractmethod
    def to_dict(self) -> Dict:
        """to_dict"""
        pass

    @abstractmethod
    def type(self) -> RequestType:
        """type"""
        pass


class VectorTopkSearchRequest(SearchRequest):
    """ TopK search request """

    def __init__(self, *,
                 vector_field: str,
                 vector: Vector = None,
                 limit: int = 50,
                 filter: str = None,
                 config: VectorSearchConfig = None):
        """init"""
        self._vector_field = vector_field
        self._vector = vector
        self._limit = limit
        self._filter = filter
        self._config = config
    
    @property
    def vector(self):
        """get vector"""
        return self._vector

    @vector.setter
    def vector(self, vector: Vector):
        """set vector"""
        self._vector = vector

    def to_dict(self):
        """to_dict"""
        res = dict()

        anns = {
            "vectorField": self._vector_field,
            "vectorFloats": self._vector.representation(),
        }

        if self._filter is not None:
            anns["filter"] = self._filter

        params = dict()
        if self._config is not None:
            params = self._config.to_dict()
        if self._limit is not None:
            params["limit"] = self._limit

        if len(params) != 0:
            anns["params"] = params

        res["anns"] = anns
        return res

    def type(self):
        return RequestType.SEARCH


class VectorRangeSearchRequest(SearchRequest):
    """ range search request """

    def __init__(self, *,
                 vector_field: str,
                 distance_range: Tuple[float, float],
                 vector: Vector = None,
                 limit: int = None,
                 filter: str = None,
                 config: VectorSearchConfig = None):
        """init"""
        self._vector_field = vector_field
        self._vector = vector
        self._distance_near = distance_range[0]
        self._distance_far = distance_range[1]
        self._limit = limit
        self._filter = filter
        self._config = config
    
    @property
    def vector(self):
        """get vector"""
        return self._vector

    @vector.setter
    def vector(self, vector: Vector):
        """set vector"""
        self._vector = vector

    def to_dict(self):
        """to_dict"""
        res = dict()

        anns = {
            "vectorField": self._vector_field,
            "vectorFloats": self._vector.representation(),
        }

        if self._filter is not None:
            anns["filter"] = self._filter

        params = dict()
        if self._config is not None:
            params = self._config.to_dict()
        if self._distance_near is not None:
            params["distanceNear"] = self._distance_near
        if self._distance_far is not None:
            params["distanceFar"] = self._distance_far
        if self._limit is not None:
            params["limit"] = self._limit

        if len(params) != 0:
            anns["params"] = params
        res["anns"] = anns

        return res

    def type(self):
        """type"""
        return RequestType.SEARCH


class VectorBatchSearchRequest(SearchRequest):
    """batch search request"""

    def __init__(self, *,
                 vector_field: str,
                 vectors: List[Vector] = None,
                 limit: int = None,
                 distance_range: Tuple[float, float] = None,
                 filter: str = None,
                 config: VectorSearchConfig = None):
        """init"""
        self._vector_field = vector_field
        self._vectors = vectors
        self._config = config
        self._limit = limit
        self._distance_range = distance_range
        self._filter = filter
    
    @property
    def vectors(self):
        """get vectors"""
        return self._vectors

    @vectors.setter
    def vectors(self, vectors: List[Vector]):
        """set vectors"""
        self._vectors = vectors

    def to_dict(self):
        """to_dict"""
        res = dict()

        vectors = []
        for vector in self._vectors:
            vectors.append(vector.representation())
        anns = {
            "vectorField": self._vector_field,
            "vectorFloats": vectors
        }
        if self._filter is not None:
            anns["filter"] = self._filter

        params = {}
        if self._config is not None:
            params = self._config.to_dict()
        if self._distance_range is not None:
            if self._distance_range[0] is not None:
                params["distanceNear"] = self._distance_range[0]
            if self._distance_range[1] is not None:
                params["distanceFar"] = self._distance_range[1]
        if self._limit is not None:
            params["limit"] = self._limit

        if len(params) != 0:
            anns["params"] = params

        res["anns"] = anns
        return res

    def type(self):
        """type"""
        return RequestType.BATCH_SEARCH


class BM25SearchRequest(SearchRequest):
    """BM25 search request"""

    def __init__(self, *,
                 index_name: str,
                 search_text: str,
                 limit: int = None,
                 filter: str = None):
        """init"""
        self._index_name = index_name
        self._search_text = search_text
        self._limit = limit
        self._filter = filter

    def to_dict(self):
        """to_dict"""
        res = {
            "BM25SearchParams": {
                "indexName": self._index_name,
                "searchText": self._search_text
            }
        }

        if self._limit is not None:
            res["limit"] = self._limit

        if self._filter is not None:
            res["filter"] = self._filter

        return res

    def type(self):
        """type"""
        return RequestType.SEARCH


VectorSearchRequest = Union[VectorTopkSearchRequest,
                            VectorRangeSearchRequest,
                            VectorBatchSearchRequest]


class HybridSearchRequest(SearchRequest):
    """Hybrid BM25 and vector search."""

    def __init__(self, *,
                 vector_request: VectorSearchRequest,
                 bm25_request: BM25SearchRequest,
                 vector_weight: float = 0.5,
                 bm25_weight: float = 0.5,
                 limit: int = None,
                 filter: str = None):
        """
        init

        'limit' and 'filter' are global settings, and they will
        apply to both vector search and BM25 search. Avoid setting
        them in 'bm25_request' or 'vector_request'.  Any settings in
        'vector_request' or 'bm25_request' for 'limit' or 'filter'
        will be overridden by the general settings.

        """

        self._vector_request = vector_request
        self._bm25_request = bm25_request
        self._vector_weight = vector_weight
        self._bm25_weight = bm25_weight
        self._limit = limit
        self._filter = filter

    def to_dict(self):
        """to_dict"""
        vector_search_params = self._vector_request.to_dict()
        vector_search_params["anns"]["weight"] = self._vector_weight

        bm25_search_params = self._bm25_request.to_dict()
        bm25_search_params["BM25SearchParams"]["weight"] = self._bm25_weight

        res = dict()
        res.update(vector_search_params)
        res.update(bm25_search_params)
        if self._limit is not None:
            res["limit"] = self._limit
        if self._filter is not None:
            res["filter"] = self._filter

        return res

    def type(self):
        """type"""
        return RequestType.SEARCH


class Table:
    """
    Table
    """
    def __init__(
            self,
            db,
            name,
            replication,
            partition,
            schema,
            enable_dynamic_field=False,
            description='',
            config=None,
            **kwargs):
        self._conn = db.conn
        self._database_name = db.database_name
        self._table_name = name
        self._replication = replication
        self._partition = partition
        self._schema = schema
        self._enable_dynamic_field = enable_dynamic_field
        self._description = description
        self._config = config
        self._create_time = kwargs.get('create_time', '')
        self._state = kwargs.get('state', None)
        self._aliases = kwargs.get('aliases', [])

    @property
    def conn(self):
        """http conn"""
        return self._conn

    @property
    def database_name(self):
        """database name"""
        return self._database_name

    @property
    def table_name(self):
        """table name"""
        return self._table_name

    @property
    def schema(self):
        """schema"""
        return self._schema

    @property
    def replication(self):
        """replication"""
        return self._replication

    @property
    def partition(self):
        """partition"""
        return self._partition

    @property
    def enable_dynamic_field(self):
        """enable dynamic field"""
        return self._enable_dynamic_field

    @property
    def description(self):
        """description"""
        return self._description

    @property
    def create_time(self):
        """create time"""
        return self._create_time

    @property
    def state(self):
        """state"""
        return self._state

    @property
    def aliases(self):
        """aliases"""
        return self._aliases

    def to_dict(self):
        """to dict"""
        res = {
            "database": self.database_name,
            "table": self.table_name,
            "description": self.description,
            "replication": self.replication,
            "partition": self.partition.to_dict(),
            "enableDynamicField": self.enable_dynamic_field,
            "schema": self.schema.to_dict(),
            "aliases": self.aliases
        }
        if self.create_time != '':
            res["createTime"] = self.create_time
        if self.state is not None:
            res["state"] = self.state
        return res

    def _merge_config(self, config):
        """merge config
        Args:
            config (dict): config need merge
        Returns:
            dict：merged config
        """
        if config is None:
            return self._config
        else:
            new_config = copy.copy(self._config)
            new_config.merge_non_none_values(config)
            return new_config

    def insert(self, rows, config=None):
        """
        insert rows
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["rows"] = []

        for row in rows:
            body['rows'].append(row.to_dict())
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.INSERT): b''},
                config=config)

    def upsert(self, rows, config=None):
        """
        upsert rows
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["rows"] = []

        for row in rows:
            body['rows'].append(row.to_dict())
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={'upsert': b''},
                config=config)

    def query(self, primary_key, partition_key=None, projections=None,
            retrieve_vector=False, read_consistency=ReadConsistency.EVENTUAL,
            config=None):
        """
        query
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["primaryKey"] = primary_key
        if partition_key is not None:
            body["partitionKey"] = partition_key
        if projections is not None:
            body["projections"] = projections
        body["retrieveVector"] = retrieve_vector
        body["readConsistency"] = read_consistency
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.QUERY): b''},
                config=config)

    def batch_query(self, keys, projections=None,
                    retrieve_vector=False,
                    read_consistency=ReadConsistency.EVENTUAL,
                    config=None):
        """
        batch_query
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["keys"] = []
        for key in keys:
            body["keys"].append(key.to_dict())

        if projections is not None:
            body["projections"] = projections
        body["retrieveVector"] = retrieve_vector
        body["readConsistency"] = read_consistency
        json_body = orjson.dumps(body)
        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST, path=uri,
                                      body=json_body,
                                      params={b'batchQuery': b''},
                                      config=config)


    @utils.deprecated("Use 'vector_search' instead.")
    def search(self, anns, partition_key=None, projections=None,
            retrieve_vector=False, read_consistency=ReadConsistency.EVENTUAL,
            config=None):
        """
        Deprecated. Use 'vector_search' instead.
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["anns"] = anns.to_dict()
        if partition_key is not None:
            body["partitionKey"] = partition_key
        if projections is not None:
            body["projections"] = projections
        body["retrieveVector"] = retrieve_vector
        body["readConsistency"] = read_consistency
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.SEARCH): b''},
                config=config)

    def vector_search(self, *,
                      request: VectorSearchRequest,
                      partition_key: Dict[str, Any] = None,
                      projections: List[str] = None,
                      read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
                      config: Dict[Any, Any] = None):
        """vector search"""
        if not isinstance(request, VectorSearchRequest.__args__):
            raise ValueError("wrong type of argument 'request'")

        return self._search(request=request,
                            partition_key=partition_key,
                            projections=projections,
                            read_consistency=read_consistency,
                            config=config)

    def bm25_search(self, *,
                    request: BM25SearchRequest,
                    partition_key: Dict[str, Any] = None,
                    projections: List[str] = None,
                    read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
                    config: Dict[Any, Any] = None):
        """BM25 search"""
        if not isinstance(request, BM25SearchRequest):
            raise ValueError("wrong type of argument 'request'")

        return self._search(request=request,
                            partition_key=partition_key,
                            projections=projections,
                            read_consistency=read_consistency,
                            config=config)

    def hybrid_search(self, *,
                      request: HybridSearchRequest,
                      partition_key: Dict[str, Any] = None,
                      projections: List[str] = None,
                      read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
                      config: Dict[Any, Any] = None):
        """hybrid search"""
        if not isinstance(request, HybridSearchRequest):
            raise ValueError("wrong type of argument 'request'")

        return self._search(request=request,
                            partition_key=partition_key,
                            projections=projections,
                            read_consistency=read_consistency,
                            config=config)

    def _search(self, *,
                request: SearchRequest,
                partition_key: Dict[str, Any] = None,
                projections: List[str] = None,
                read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
                config: Dict[Any, Any] = None):
        """internal use only"""
        if not self.conn:
            raise ClientError('conn is closed')

        body = request.to_dict()
        body["database"] = self.database_name
        body["table"] = self.table_name

        if partition_key is not None:
            body["partitionKey"] = partition_key
        if projections is not None:
            body["projections"] = projections
        body["readConsistency"] = read_consistency
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        req_type = bytes(request.type())
        return self.conn.send_request(http_methods.POST,
                                      path=uri,
                                      body=json_body,
                                      params={req_type: b''},
                                      config=config)

    def delete(self, primary_key=None, partition_key=None, filter=None, config=None):
        """
        delete row
        """
        if not self.conn:
            raise ClientError('conn is closed')

        if primary_key is None and filter is None:
            raise ValueError('requiring primary_key or filter')
        if primary_key is not None and filter is not None:
            raise ValueError('only one of primary_key and filter should exist')
        if partition_key is not None and filter is not None:
            raise ValueError('only one of partition_key and filter should exist')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        if primary_key is not None:
            body["primaryKey"] = primary_key
        if partition_key is not None:
            body["partitionKey"] = partition_key
        if filter is not None:
            body["filter"] = filter
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.DELETE): b''},
                config=config)

    def update(self, primary_key=None, partition_key=None, update_fields=None, config=None):
        """
        update row
        """
        if not self.conn:
            raise ClientError('conn is closed')

        if primary_key is None and update_fields is None:
            raise ValueError('requiring primary_key and update_fields')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        if primary_key is not None:
            body["primaryKey"] = primary_key
        if partition_key is not None:
            body["partitionKey"] = partition_key
        if update_fields is not None:
            body["update"] = update_fields
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.UPDATE): b''},
                config=config)

    def select(self, filter=None, marker=None, projections=None, read_consistency=ReadConsistency.EVENTUAL, limit=10,
            config=None):
        """
        select
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["readConsistency"] = read_consistency
        body["limit"] = limit
        if filter is not None:
            body["filter"] = filter
        if marker is not None:
            body["marker"] = marker
        if projections is not None:
            body["projections"] = projections
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.SELECT): b''},
                config=config)

    @utils.deprecated("Use vector_search() with VectorBatchSearchRequest instead")
    def batch_search(self, anns, partition_key=None, projections=None, 
            retrieve_vector=False, read_consistency=ReadConsistency.EVENTUAL, 
            config=None):
        """
        Deprecated. Use vector_search() with VectorBatchSearchRequest instead.
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["anns"] = anns.to_dict()
        if partition_key is not None:
            body["partitionKey"] = partition_key
        if projections is not None:
            body["projections"] = projections
        body["retrieveVector"] = retrieve_vector
        body["readConsistency"] = read_consistency
        json_body = orjson.dumps(body)
        
        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'row')
        
        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.BATCH_SEARCH): b''},
                config=config)
    
    def add_fields(self, schema, config=None):
        """
        add_fields
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["schema"] = schema.to_dict()
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'table')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.ADD_FIELD): b''},
                config=config)

    def create_indexes(self, indexes, config=None):
        """
        create indexes
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["indexes"] = []

        for index in indexes:
            if isinstance(index, VectorIndex):
                body["indexes"].append(index.to_dict())
            else:
                raise ClientError("not supported index type")

        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'index')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.CREATE): b''},
                config=config)

    def modify_index(self, index_name, auto_build, auto_build_index_policy=DefaultAutoBuildPolicy, config=None):
        """
        modify index
        """
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        body["index"] = {
            "indexName": index_name,
            "autoBuild": auto_build
        }
        if auto_build:
            body["index"]["autoBuildPolicy"] = auto_build_index_policy.to_dict()
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'index')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.MODIFY): b''},
                config=config)

    def drop_index(self, index_name, config=None):
        """drop index"""
        if not self.conn:
            raise ClientError('conn is closed')

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'index')
        return self.conn.send_request(http_methods.DELETE,
                path=uri,
                params={
                    b'database': self.database_name,
                    b'table': self.table_name,
                    b'indexName': index_name},
                config=config)

    def rebuild_index(self, index_name, config=None):
        """build vector index"""
        if not self.conn:
            raise ClientError('conn is closed')

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'index')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=orjson.dumps({"database": self.database_name,
                    "table": self.table_name,
                    "indexName": index_name}),
                params={bytes(RequestType.REBUILD): b''},
                config=config)

    def describe_index(self, index_name, config=None):
        """describe index"""
        if not self.conn:
            raise ClientError('conn is closed')

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'index')

        response = self.conn.send_request(http_methods.POST,
                path=uri,
                params={bytes(RequestType.DESC): b''},
                body=orjson.dumps({
                    'database': self.database_name,
                    'table': self.table_name,
                    'indexName': index_name
                }),
                config=config)
        index = response.index
        auto_build_index_policy = None
        if "autoBuildPolicy" in index:
            auto_build_index_policy = AutoBuildTool.get_auto_build_index_policy(index["autoBuildPolicy"])
        if index["indexType"] == IndexType.HNSW.value:
            return VectorIndex(
                index_name=index["indexName"],
                index_type=IndexType.HNSW,
                field=index["field"],
                metric_type=getattr(MetricType, index["metricType"], None),
                params=HNSWParams(m=index["params"]["M"],
                    efconstruction=index["params"]["efConstruction"]),
                auto_build=index["autoBuild"],
                auto_build_index_policy=auto_build_index_policy,
                state=getattr(IndexState, index["state"], None))
        elif index["indexType"] == IndexType.HNSWPQ.value:
            return VectorIndex(
                index_name=index["indexName"],
                index_type=IndexType.HNSWPQ,
                field=index["field"],
                metric_type=getattr(MetricType, index["metricType"], None),
                params=HNSWPQParams(m=index["params"]["M"],
                    efconstruction=index["params"]["efConstruction"],
                    NSQ=index["params"]["NSQ"],
                    samplerate=index["params"]["sampleRate"]),
                auto_build=index["autoBuild"],
                auto_build_index_policy=auto_build_index_policy,
                state=getattr(IndexState, index["state"], None))
        elif index["indexType"] == IndexType.FLAT.value:
            return VectorIndex(
                index_name=index["indexName"],
                index_type=IndexType.FLAT,
                field=index["field"],
                metric_type=getattr(MetricType, index["metricType"], None),
                auto_build=index["autoBuild"],
                auto_build_index_policy=auto_build_index_policy,
                state=getattr(IndexState, index["state"], None))
        elif index["indexType"] == IndexType.PUCK.value:
            return VectorIndex(
                index_name=index["indexName"],
                index_type=IndexType.PUCK,
                field=index["field"],
                metric_type=getattr(MetricType, index["metricType"], None),
                params=PUCKParams(coarseClusterCount=index["params"]["coarseClusterCount"],
                        fineClusterCount=index["params"]["fineClusterCount"]),
                auto_build=index["autoBuild"],
                auto_build_index_policy=auto_build_index_policy,
                state=getattr(IndexState, index["state"], None))
        elif index["indexType"] == IndexType.SECONDARY_INDEX.value:
            return SecondaryIndex(
                index_name=index["indexName"],
                field=index["field"])
        elif index["indexType"] == IndexType.INVERTED_INDEX.value:
            return InvertedIndex(
                index_name=index["indexName"],
                fields=index["fields"],
                params=InvertedIndexParams(analyzer=getattr(InvertedIndexAnalyzer, index["params"]["analyzer"], None),
                                    parse_mode=getattr(InvertedIndexParseMode, index["params"]["parseMode"], None)))
        else:
            raise ClientError("not supported index type:%s" % (index["indexType"]))


    def stats(self, config=None):
        """show table stats"""
        if not self.conn:
            raise ClientError('conn is closed')

        body = {}
        body["database"] = self.database_name
        body["table"] = self.table_name
        json_body = orjson.dumps(body)

        config = self._merge_config(config)
        uri = utils.append_uri(client.URL_PREFIX, client.URL_VERSION, 'table')

        return self.conn.send_request(http_methods.POST,
                path=uri,
                body=json_body,
                params={bytes(RequestType.STATS): b''},
                config=config)


class Row:
    """
    row, the object for document insert, query and search, the parameter depends on
    the schema of table.
    """

    def __init__(self, **kwargs) -> None:
        self._data = kwargs

    def to_dict(self):
        """to dict"""
        return self._data


@utils.deprecated("Use VectorSearchConfig instead")
class HNSWSearchParams:
    "hnsw search params"

    def __init__(self, ef=None, distance_far=None, distance_near=None, limit=50,
            pruning=True):
        self._ef = ef
        self._distance_far = distance_far
        self._distance_near = distance_near
        self._limit = limit
        self._pruning = pruning

    def to_dict(self):
        """to dict"""
        res = {}
        if self._ef is not None:
            res['ef'] = self._ef
        if self._distance_far is not None:
            res['distanceFar'] = self._distance_far
        if self._distance_near is not None:
            res['distanceNear'] = self._distance_near
        res['limit'] = self._limit
        res['pruning'] = self._pruning
        return res

class HNSWPQSearchParams:
    "hnswpq search params"

    def __init__(self, ef=None, distance_far=None, distance_near=None, limit=50):
        self._ef = ef
        self._distance_far = distance_far
        self._distance_near = distance_near
        self._limit = limit

    def to_dict(self):
        """to dict"""
        res = {}
        if self._ef is not None:
            res['ef'] = self._ef
        if self._distance_far is not None:
            res['distanceFar'] = self._distance_far
        if self._distance_near is not None:
            res['distanceNear'] = self._distance_near
        res['limit'] = self._limit
        return res


@utils.deprecated("Use VectorSearchConfig instead")
class HNSWPQSearchParams:
    "hnswpq search params"

    def __init__(self, ef=None, distance_far=None, distance_near=None, limit=50):
        self._ef = ef
        self._distance_far = distance_far
        self._distance_near = distance_near
        self._limit = limit

    def to_dict(self):
        """to dict"""
        res = {}
        if self._ef is not None:
            res['ef'] = self._ef
        if self._distance_far is not None:
            res['distanceFar'] = self._distance_far
        if self._distance_near is not None:
            res['distanceNear'] = self._distance_near
        res['limit'] = self._limit
        return res


@utils.deprecated("Use VectorSearchConfig instead")
class PUCKSearchParams:
    "puck search params"

    def __init__(self, searchCoarseCount, limit=50) -> None:
        self._limit = limit
        self._searchCoarseCount = searchCoarseCount

    def to_dict(self):
        """to dict"""
        res = {}

        res['searchCoarseCount'] = self._searchCoarseCount
        res['limit'] = self._limit

        return res


@utils.deprecated("Use VectorSearchConfig instead")
class FLATSearchParams:
    "flat search params"

    def __init__(self, distance_far=None, distance_near=None, limit=50):
        self._distance_far = distance_far
        self._distance_near = distance_near
        self._limit = limit

    def to_dict(self):
        """to dict"""
        res = {}
        if self._distance_far is not None:
            res['distanceFar'] = self._distance_far
        if self._distance_near is not None:
            res['distanceNear'] = self._distance_near
        res['limit'] = self._limit
        return res



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provide appbuilder implement to split doc.
"""
import logging
import uuid
import time
from typing import List
from pymochow.ai.splitter import DocSplitter
from pymochow.model.document import DocumentChunk
from pymochow.model.enum import (
    DocSplitMode,
    Lang
)
from langchain.text_splitter import RecursiveCharacterTextSplitter

_logger = logging.getLogger(__name__)

class LangchainDocSplitter(DocSplitter):
    """
    LangchainDocSplitter: A concrete implementation of the DocProcessor class. 
    This class is responsible to parser and split a document into chunks based on a given splitting mode.
    """

    def __init__(
        self, 
        split_mode=DocSplitMode.PAGE,
        maximum_page_length=800,
        page_overlap_length=200,
        maximum_pages_totake=None
    ):
        """
        Initialize the QianfanDocProcessor.
        split_mode : DocSplitMode
            The mode of splitting the document (e.g., by page). The default is to split by pages.
        
        maximum_page_length : int, optional
            The maximum length of a chunk in terms of 
            pages (or another unit depending on the split mode). Defaults to None.
        
        page_overlap_length : int, optional
            The number of overlapping units (pages, words, etc.) between consecutive chunks, 
            ensuring some content overlap. Defaults to None.
        
        maximum_pages_totake : int, optional
            The total number of pages (or another unit) to process from the document. 
            This acts as a limit on how much of the document is parsed. Defaults to None.
        """
        self._split_mode = split_mode
        self._maximum_page_length = maximum_page_length
        self._page_overlap_length = page_overlap_length
        self._maximum_pages_totake = maximum_pages_totake

    def split(self, chunk) -> List[DocumentChunk]:
        """
        split the document into chunks based on the provided parameters.
        
        Parameters:
        ----------
        chunk : DocumentChunk
            The chunk to be parsed and split into chunks.
        
        Returns:
        -------
        List[DocumentChunk]
            A list of DocumentChunk objects, where each chunk represents a part of the parsed document.
        """
        if self._split_mode != DocSplitMode.PAGE:
            raise ValueError(f"Unsupported split mode: {self._split_mode}. Only DocSplitMode.PAGE is supported.")
        
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=self._maximum_page_length,  # Maximum length of each chunk
            chunk_overlap=self._page_overlap_length  # Overlap between chunks
        )

        chunks = text_splitter.split_text(chunk.content)

        document_chunks = []
        for idx, content in enumerate(chunks):
            new_chunk = DocumentChunk(
                kb_id=chunk.kb_id,
                doc_id=chunk.doc_id,
                chunk_id=str(uuid.uuid4()),  # Generate a unique chunk ID
                doc_name=chunk.doc_name,
                sequence_number=idx + 1,  # Create a sequence number for the chunk
                content=content,
                content_len=len(content),
                ctime=int(time.time())  # Use current time for chunk creation time
            )
            document_chunks.append(new_chunk)
            
            if self._maximum_pages_totake is not None and len(chunks) >= self._maximum_pages_totake:
                break

        return document_chunks



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide local document hub.
"""
from pymochow.model.document import Document
from pymochow.ai.dochub import DocumentHub, DocumentHubEnv
import logging
import os
import shutil
from typing import List

_logger = logging.getLogger(__name__)

class LocalDocumentHub(DocumentHub):
    """
    LocalDocumentHub: A concrete implementation of DocumentHub that manages documents locally.
    """

    def __init__(self, env: DocumentHubEnv):
        """
        Initialize the LocalDocumentHub with the environment configuration.

        Parameters:
        env (DocumentHubEnv): The environment configuration containing the local cache path.
        """
        self._env = env
        self._root_path = env.root_path()
        self._schema = "local://"
        if self._root_path.startswith(self._schema):
            # Strip the 'local://' prefix for subsequent path operations
            self._root_path = self._root_path[len(self._schema):]

        else:
            raise ValueError(f"Invalid root_path format: {self._root_path}. Expected 'local://xxx/xxx'.")
        
        if not self._root_path.startswith("/"):
            self._root_path = "/" + self._root_path

        # Ensure the local cache directory exists
        if not os.path.exists(self._root_path):
            os.makedirs(self._root_path)

    def add(self, doc: Document) -> Document:
        """
        Add a document to the hub.

        Parameters:
        doc (Document): The document to be added to hub.
        """
        if doc.doc_name is None or doc.file_path is None:
            raise ValueError("Document file path or name is not set.")
        # Combine cache path with the document's directory structure
        target_path = os.path.join(self._root_path, doc.doc_name)

        # Ensure the target directory exists
        target_dir = os.path.dirname(target_path)
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        
        # Copy the document to the target path
        shutil.copy(doc.file_path, target_path)
        
        _, file_extension = os.path.splitext(doc.doc_name)
        doc.doc_type = file_extension.lstrip(".")  # Remove the dot from extension

        # Update the document's file path, size, and creation time
        doc.file_path = target_path
        doc.uri = self._schema + target_path
        doc.size = os.path.getsize(target_path)  # Get file size
        doc.ctime = int(os.path.getctime(target_path))  # Get creation time
        _logger.info(f"Document '{doc.doc_name}' has been added to hub. path: {target_dir}.")
        return doc
        
    def remove(self, doc: Document):
        """
        Remove a document from the hub.

        Parameters:
        doc (Document): The document to be removed from hub.
        """
        target_path = os.path.join(self._root_path, doc.doc_name)
        if os.path.exists(target_path):
            os.remove(target_path)
            _logger.info(f"Document '{doc.doc_name}' has been removed from hub.")

            target_dir = os.path.dirname(target_path)
            
            while target_dir != self._root_path:
                if os.path.isdir(target_dir) and not os.listdir(target_dir):  # Directory is empty
                    os.rmdir(target_dir)
                    _logger.info(f"Directory '{target_dir}' has been removed as it is empty.")
                    target_dir = os.path.dirname(target_dir)  # Move up to the next parent directory
                else:
                    break
        else:
            _logger.info(f"Document '{doc.doc_name}' does not exist in hub.")

    def list(self) -> List[Document]:
        """
        List all documents in hub.

        Returns:
        List[Document]: A list of all documents in the hub.
        """
        documents = []
        for root, dirs, files in os.walk(self._root_path):
            for filename in files:
                file_path = os.path.join(root, filename)
                relative_path = os.path.relpath(file_path, self._root_path)
                # Placeholder values for doc attributes that are not stored in the file system.
                _, file_extension = os.path.splitext(filename)
                doc_type = file_extension.lstrip(".")
                
                doc = Document(
                    doc_name=relative_path, 
                    doc_type=doc_type,
                    file_path=file_path, 
                    uri=self._schema + file_path,
                    size=os.path.getsize(file_path),
                    ctime=int(os.path.getctime(file_path))
                )
                documents.append(doc)
        return documents

    def load(self, doc: Document) -> Document:
        """
        Load a document if it in hub.

        Parameters:
        doc (Document): The document to load.

        Returns:
        doc (Document): If the document in hub then load it.
        """
        target_path = os.path.join(self._root_path, doc.doc_name)
        
        if os.path.exists(target_path):
            # Load the document's metadata (size and ctime) from the cached file
            doc.file_path = target_path
            doc.uri = self._schema + target_path
            doc.size = os.path.getsize(target_path)
            doc.ctime = int(os.path.getctime(target_path))
            
            _, file_extension = os.path.splitext(doc.doc_name)
            doc.doc_type = file_extension.lstrip(".")

            _logger.info(f"Document '{doc.doc_name}' loaded from hub.")
            return doc
        else:
            # If the document is not cached, return None or raise an error
            _logger.info(f"Document '{doc.doc_name}' is not in hub.")
            return None



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provide appbuilder implement to parser doc.
"""

import logging
import time
import os
import uuid
from typing import List
from pymochow.ai.parser import DocParser
from pymochow.model.document import DocumentChunk
from langchain_community.document_loaders import PyPDFLoader

_logger = logging.getLogger(__name__)

class LangchainDocParser(DocParser):
    """
    LangchainDocParser: A concrete implementation of the DocProcessor class. 
    This class is responsible to parser a document into chunks based on a given splitting mode.
    """

    def __init__(
        self
    ):
        """
        Initialize the QianfanDocProcessor.
        """
        pass

    def parse(self, doc) -> DocumentChunk:
        """
        Parse the document into chunk based on the provided parameters.
        
        Parameters:
        ----------
        doc : Document
            The document to be parsed and split into chunks.
        
        Returns:
        -------
        DocumentChunk
            A  DocumentChunk object
        """
        if doc.file_path is None:
            raise ValueError("Document does not have a file path.")
        
        file_extension = os.path.splitext(doc.file_path)[1].lower()
        if file_extension != '.pdf':
            raise ValueError("Only PDF files are supported. Provided file is not a PDF.")

        # Using PyPDFLoader to load the document (specifically for PDF files)
        loader = PyPDFLoader(doc.file_path)
        langchain_docs = loader.load()

        document_content = "\n".join([doc.page_content for doc in langchain_docs])

        chunk_obj = DocumentChunk(
            kb_id=doc.kb_id,
            doc_id=doc.doc_id,
            chunk_id=str(uuid.uuid4()),  # Generate a unique chunk ID
            doc_name=doc.doc_name,
            sequence_number=1,  # Since it's a single chunk, the sequence number is 1
            content=document_content,
            content_len=len(document_content),  # Set the length of the chunk content
            ctime=int(time.time())  # Use current time for chunk creation time
        )
        return chunk_obj



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provide interface for pipeline.
"""
from abc import ABC, abstractmethod
from typing import Any, Dict, List
from pymochow.ai.embedder import Embedder
from pymochow.model.table import (
    Table,
    VectorSearchRequest,
    BM25SearchRequest,
    HybridSearchRequest
)
from pymochow.model.enum import ReadConsistency

class Pipeline(ABC):
    """
    Pipeline: An abstract base class that defines the interface for ingesting
    doc into vectordb and search from it.
    """
    @abstractmethod
    def ingest_doc(doc, 
            doc_processor=None, 
            embedder=None, 
            meta_table=None,
            doc_to_row_mapping=None,
            chunk_table=None,
            chunk_to_row_mapping=None):
        """
        Abstract method for processing and storing the ingestion of documents and their chunks.
        
        Parameters:
            doc (Document): The document object to be ingested.
            doc_processor (Processor, optional): A tool used to parse and split document.
            embedder (Embedder, optional): A tool used to generate embeddings of the document content.
            meta_table (Table, optional): The table in the database that stores document data.
            doc_to_row_mapping (dict, optional): A JSON-like dictionary that defines 
                the mapping between document object attributes and database table columns.
                Example mapping:
                {
                    'doc_id': 'document_id',  # Maps 'doc_id' in the document to 'document_id' in the database
                    'doc_name': 'document_name'  # Maps 'doc_name' in the document to 'document_name' in the database
                    # Add more mappings as needed...
                }
            chunk_table (Table, optional): The table in the database used to store document chunks, 
                if the document is processed in chunks.
            chunk_to_row_mapping (dict, optional): Similar to doc_to_row_mapping, 
                this dictionary defines how attributes of chunks map to database table columns.
                Example:
                {
                    'chunk_id': 'chunk_id',  # Example mapping, no change
                    # Additional mappings can be added here...
                }

        Returns:
            None: This method does not return anything but may modify data in the database or other storage systems.

        Note:
            This is an abstract method that must be implemented in any subclass inheriting 
             from Pipeline with specific logic for ingestion.
        """
        pass

    def vector_search(
        self, 
        search_contents: List[str], 
        embedder: Embedder, 
        table: Table, 
        search_request: VectorSearchRequest,
        partition_key: Dict[str, Any] = None,
        projections: List[str] = None,
        read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
        config: Dict[Any, Any] = None
    ):
        """
        Perform a vector-based search operation.

        This method converts the search content into a vector using the embedder and performs a
        vector search in the specified table, returning the most similar results to the query.

        Parameters:
        ----------
        search_contents : List[str]
            The input search content, usually in the form of text or already embedded vector.
        embedder : Embedder
            The embedder object used to convert the search content into vector form.
        table : Table
            The target table where the search is conducted.
        search_request : VectorSearchRequest
            The search request object containing parameters like TopK, filters, etc.
        partition_key : Dict[str, Any], optional
            The partition key to narrow the search to a specific partition, default is None.
        projections : List[str], optional
            The list of fields to include in the search result, default is None.
        read_consistency : ReadConsistency, optional
            The level of read consistency required, default is EVENTUAL.
        config : Dict[Any, Any], optional
            Additional configurations for the search, default is None.
        """
        pass

    def bm25_search(
        self,
        table: Table,
        search_request: BM25SearchRequest,
        partition_key: Dict[str, Any] = None,
        projections: List[str] = None,
        read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
        config: Dict[Any, Any] = None
    ):
        """
        Perform a BM25-based text search operation.

        This method uses the BM25 search algorithm to perform a text-based search on the specified
        table, returning the most relevant documents based on the query terms.

        Parameters:
        ----------
        table : Table
            The target table where the search will be performed.
        search_request : BM25SearchRequest
            The search request object, which contains query terms and other parameters.
        partition_key : Dict[str, Any], optional
            The partition key to narrow the search to a specific partition, default is None.
        projections : List[str], optional
            The list of fields to include in the search result, default is None.
        read_consistency : ReadConsistency, optional
            The level of read consistency required, default is EVENTUAL.
        config : Dict[Any, Any], optional
            Additional configurations for the search, default is None.
        """
        pass

    def hybrid_search(
        self, 
        search_contents: List[str], 
        embedder: Embedder, 
        table: Table,
        search_request: HybridSearchRequest,
        partition_key: Dict[str, Any] = None,
        projections: List[str] = None,
        read_consistency: ReadConsistency = ReadConsistency.EVENTUAL,
        config: Dict[Any, Any] = None
    ):
        """
        Perform a hybrid search (vector + traditional text search).

        This method combines vector search with BM25 text search, suitable for scenarios
        requiring both semantic and keyword-based search.

        Parameters:
        ----------
        search_contents : List[str]
            The input search content, usually text-based.
        embedder : Embedder
            The embedder object used to convert the search content into vectors.
        table : Table
            The target table where the search is conducted.
        partition_key : Dict[str, Any], optional
            The partition key to narrow the search to a specific partition, default is None.
        projections : List[str], optional
            The list of fields to include in the search result, default is None.
        read_consistency : ReadConsistency, optional
            The level of read consistency required, default is EVENTUAL.
        config : Dict[Any, Any], optional
            Additional configurations for the search, default is None.
        """
        pass



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide document model.
"""
import uuid
import time
from pymochow.model.table import Row
from pymochow.model.enum import (
    DocumentLayout,
    Lang
)

class Document:
    """Document Model"""
   
    def __init__(
        self,
        kb_id=None,
        doc_id=None,
        doc_name=None,
        doc_type=None,
        layout=DocumentLayout.GENERAL,
        lang=Lang.ZH,
        file_path=None,
        uri=None,
        size=None,
        ctime=None
    ):
        if not doc_name:
            raise ValueError("doc_name is required")  # Raise an error if doc_name is not provided
        
        self.doc_id = doc_id if doc_id is not None else str(uuid.uuid4())
        
        self.kb_id = kb_id
        self.doc_name = doc_name
        self.doc_type = doc_type # pdf/doc/markdown/excel
        self.layout = layout # Q&A/paper/law
        self.lang = lang # en/zh
        self.file_path = file_path # local file path
        self.uri = uri # uri in FileSystem, example: bos://bucket/xxx
        self.size = size
        self.ctime = int(ctime if ctime is not None else time.time())
    
    def to_row(self, field_mapping=None):
        """
        Convert Document to Row with custom field mapping.

        Parameters:
        ----------
        field_mapping: dict
            A dictionary mapping Document attributes to Row fields.

        Returns:
        -------
        Row
            A Row object created from the Document object with custom field names.
        """
        data = {}
        for field, value in self.__dict__.items():
            if value is None:
                continue

            # If field_mapping is None, convert all fields
            if field_mapping is None:
                new_field_name = field  # Keep original field name
            # If field_mapping exists, only convert fields that are in the mapping
            elif field in field_mapping:
                new_field_name = field_mapping.get(field, field)
            else:
                continue  # Skip fields that are not in field_mapping
            data[new_field_name] = value

        return Row(**data)

    def to_dict(self):
        """
        Convert the Document object to a dictionary representation.

        Returns:
        -------
        dict
            A dictionary representation of the Document object.
        """
        # Simply return the __dict__ attribute, which is a dictionary of all instance attributes
        return self.__dict__.copy()

class DocumentChunk:
    """Document Chunk Model"""

    def __init__(
        self, 
        kb_id=None,
        doc_id=None,
        chunk_id=None, 
        doc_name=None,
        sequence_number=None,
        content=None,
        content_len=None,
        embedding=None,
        ctime=None,
        **kwargs
    ):
        
        if not kb_id:
            raise ValueError("kb_id cannot be None")
        if not doc_id:
            raise ValueError("doc_id cannot be None")
        if not doc_name:
            raise ValueError("doc_name cannot be None")

        # If chunk_id is None, generate a UUID
        self.chunk_id = chunk_id if chunk_id is not None else str(uuid.uuid4())
        
        self.kb_id = kb_id
        self.doc_id = doc_id
        self.doc_name = doc_name
        self.sequence_number = sequence_number
        self.content = content
        self.content_len = content_len
        self.embedding = embedding
        self.ctime = int(ctime if ctime is not None else time.time())

        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_row(self, field_mapping=None):
        """Convert DocumentChunk to Row with custom field mapping
        field_mapping = {
            'doc_id': 'document_id',  # example: doc_id -> document_id
            'chunk_id': 'chunk_id',   # no change, just for example
            'doc_name': 'document_name',  # doc_name -> document_name
            # Add more mappings as needed...
        }
        """
        data = {}
        for field, value in self.__dict__.items():
            if value is None:
                continue

            # If field_mapping is None, convert all fields
            if field_mapping is None:
                new_field_name = field  # Keep original field name
            # If field_mapping exists, only convert fields that are in the mapping
            elif field in field_mapping:
                new_field_name = field_mapping.get(field, field)
            else:
                continue  # Skip fields that are not in field_mapping
            data[new_field_name] = value

        return Row(**data)
    
    def to_dict(self):
        """
        Convert the DocumentChunk object to a dictionary representation.

        Returns:
        -------
        dict
            A dictionary representation of the DocumentChunk object.
        """
        # Simply return the __dict__ attribute, which is a dictionary of all instance attributes
        return self.__dict__.copy()



# Copyright 2023 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide enum.
"""
from enum import Enum, unique


@unique
class MetricType(Enum):
    """
    The metric type of the vector index.
    """
    L2 = "L2"
    IP = "IP"
    COSINE = "COSINE"


@unique
class IndexType(Enum):
    """index type"""
    # vector index type
    HNSW = "HNSW"
    FLAT = "FLAT"
    PUCK = "PUCK"
    HNSWPQ = "HNSWPQ"
    
    # scalar index type
    SECONDARY_INDEX = "SECONDARY"

    # inverted index type
    INVERTED_INDEX = "INVERTED"


@unique
class InvertedIndexAnalyzer(Enum):
    """inverted index analyzer"""
    ENGLISH_ANALYZER = "ENGLISH_ANALYZER"
    CHINESE_ANALYZER = "CHINESE_ANALYZER"
    DEFAULT_ANALYZER = "DEFAULT_ANALYZER"


@unique
class InvertedIndexParseMode(Enum):
    """parse mode"""
    COARSE_MODE = "COARSE_MODE"
    FINE_MODE = "FINE_MODE"


class InvertedIndexFieldAttribute(Enum):
    """InvertedIndexFieldsAttribute"""
    NOT_ANALYZED = "ATTRIBUTE_NOT_ANALYZED"
    ANALYZED = "ATTRIBUTE_ANALYZED"


@unique
class FieldType(Enum):
    """
    Field type
    """
    # scalar field type
    BOOL = "BOOL"
    INT8 = "INT8"
    UINT8 = "UINT8"
    INT16 = "INT16"
    UINT16 = "UINT16"
    INT32 = "INT32"
    UINT32 = "UINT32"
    INT64 = "INT64"
    UINT64 = "UINT64"
    FLOAT = "FLOAT"
    DOUBLE = "DOUBLE"
    DATE = "DATE"
    DATETIME = "DATETIME"
    TIMESTAMP = "TIMESTAMP"
    STRING = "STRING"
    BINARY = "BINARY"
    UUID = "UUID"
    TEXT = "TEXT"
    TEXT_GBK = "TEXT_GBK"
    TEXT_GB18030 = "TEXT_GB18030"

    # vector field type
    FLOAT_VECTOR = "FLOAT_VECTOR"


@unique
class PartitionType(Enum):
    """
    Partition Type
    """
    HASH = "HASH"


@unique
class ServerErrCode(Enum):
    """
    Server error no
    """
    INTERNAL_ERROR = 1
    INVALID_PARAMETER = 2

    INVALID_HTTP_URL = 10
    INVALID_HTTP_HEADER = 11
    INVALID_HTTP_BODY = 12
    MISS_SSL_CERTIFICATES = 13

    USER_NOT_EXIST = 20
    USER_ALREADY_EXIST = 21
    ROLE_NOT_EXIST = 22
    ROLE_ALREADY_EXIST = 23
    AUTHENTICATION_FAILED = 24
    PERMISSION_DENIED = 25

    # Database errors
    DB_NOT_EXIST = 50
    DB_ALREADY_EXIST = 51
    DB_TOO_MANY_TABLES = 52
    DB_NOT_EMPTY = 53

    # Table errors
    INVALID_TABLE_SCHEMA = 60
    INVALID_PARTITION_PARAMETERS = 61
    TABLE_TOO_MANY_FIELDS = 62
    TABLE_TOO_MANY_FAMILIES = 63
    TABLE_TOO_MANY_PRIMARY_KEYS = 64
    TABLE_TOO_MANY_PARTITION_KEYS = 65
    TABLE_TOO_MANY_VECTOR_FIELDS = 66
    TABLE_TOO_MANY_INDEXES = 67
    DYNAMIC_SCHEMA_ERROR = 68
    TABLE_NOT_EXIST = 69
    TABLE_ALREADY_EXIST = 70
    INVALID_TABLE_STATE = 71
    TABLE_NOT_READY = 72
    ALIAS_NOT_EXIST = 73
    ALIAS_ALREADY_EXIST = 74

    # Field errors
    FIELD_NOT_EXIST = 80
    FIELD_ALREADY_EXIST = 81
    VECTOR_FIELD_NOT_EXIST = 82

    # Index errors
    INVALID_INDEX_SCHEMA = 90
    INDEX_NOT_EXIST = 91
    INDEX_ALREADY_EXIST = 92
    INDEX_DUPLICATED = 93
    INVALID_INDEX_STATE = 94

    # Row errors
    PRIMARY_KEY_DUPLICATED = 100
    ROW_KEY_NOT_FOUND = 101


@unique
class TableState(Enum):
    """
    Table State
    """
    CREATING = "CREATING"
    NORMAL = "NORMAL"
    DELETING = "DELETING"


@unique
class ReadConsistency(Enum):
    """
    Read Consistency
    """
    EVENTUAL = "EVENTUAL"
    STRONG = "STRONG"


@unique
class IndexState(Enum):
    """
    Index State
    """
    BUILDING = "BUILDING"
    NORMAL = "NORMAL"


@unique
class AutoBuildPolicyType(Enum):
    """
    Auto Build Policy Type
    """
    def __str__(self):
        """
            将对象转换为字符串，返回值为对象的value属性。
        如果value属性不存在或者是None，则返回空字符串。

        Returns:
            str -- 对象的value属性，如果不存在或者是None，则返回空字符串。
        """
        return str(self.value)
    TIMING = "TIMING"
    PERIODICAL = "PERIODICAL"
    ROW_COUNT_INCREMENT = "ROW_COUNT_INCREMENT"
    #RESOURCE_USAGE = 4


@unique
class RequestType(Enum):
    """request type"""

    def __str__(self):
        return str(self.value)

    def __bytes__(self):
        return bytes(self.value, 'utf-8')

    SHOW = "show"
    LIST = "list"
    DESC = "desc"
    ADD_FIELD = "addField"
    MODIFY = "modify"
    REBUILD = "rebuild"
    INSERT = "insert"
    UPSERT = "upsert"
    DELETE = "delete"
    UPDATE = "update"
    QUERY = "query"
    BATCH_QUERY = "batchQuery"
    SEARCH = "search"
    BATCH_SEARCH = "batchSearch"
    CREATE = "create"
    DROP = "drop"
    ALIAS = "alias"
    UNALIAS = "unalias"
    STATS = "stats"
    SELECT = "select"


@unique
class DocumentLayout(Enum):
    """
    DocumentLayout: Enum representing the layout or structure of a document.

    GENERAL: Represents a general layout, not specific to any particular format.
    """
    GENERAL = "GENERAL"


@unique
class Lang(Enum):
    """
    Lang: Enum representing the language of a document.

    ZH: Represents Chinese language (ZH).
    EN: Represents English language (EN).
    """
    ZH = "ZH"
    EN = "EN"


@unique
class DocSplitMode(Enum):
    PAGE = "PAGE"
    SENTENCE = "SENTENCE"
